//@ts-check
"use strict";

/**
@typedef TileCollision
@type {{
  rect: PIXI.Rectangle;
  ignoreFlyers: boolean;
}}
*/

/**
@param {number} x
@param {number} y
@returns {TileCollision}
*/
function getTileCollisionAt(x, y) {
  const i = Math.floor(y) * inGame.activeWorld.width + Math.floor(x);

  for (let layerIndex = 0; layerIndex < inGame.activeWorld.layers.length; layerIndex++) {
    const layer = inGame.activeWorld.layers[layerIndex];

    const tile = layer.data[i];

    const tileId = (tile & ~(horizontalFlipFlag | verticalFlipFlag | diagonalFlipFlag)) - 1;
    const baseTileCollision = inGame.tileset.collisionsById[tileId];
    if (baseTileCollision == null) continue;

    const orientedTileCollision = new PIXI.Rectangle(
      baseTileCollision.x / tileSize,
      baseTileCollision.y / tileSize,
      baseTileCollision.width / tileSize,
      baseTileCollision.height / tileSize);

    const horizontalFlip = (tile & horizontalFlipFlag) !== 0;
    const verticalFlip = (tile & verticalFlipFlag) !== 0;
    const diagonalFlip = (tile & diagonalFlipFlag) !== 0;

    if (diagonalFlip) {
      [orientedTileCollision.x, orientedTileCollision.y] = [orientedTileCollision.y, orientedTileCollision.x];
      [orientedTileCollision.width, orientedTileCollision.height] = [orientedTileCollision.height, orientedTileCollision.width];
    }

    if (horizontalFlip) orientedTileCollision.x = 1 - orientedTileCollision.width;
    if (verticalFlip) orientedTileCollision.y = 1 - orientedTileCollision.height;

    orientedTileCollision.x += Math.floor(x);
    orientedTileCollision.y += Math.floor(y);

    return { rect: orientedTileCollision, ignoreFlyers: baseTileCollision.ignoreFlyers };
  }

  return null;
}

/**
 * @typedef {Object} EntityLike
 * @type {{
 * pos: PIXI.Point,
 * collideBox: {width: number, height: number},
 * hitbox: {width: number, height: number},
 * speed: number }}
 */

/**
 * @param {EntityLike} entity
 * @param {{ x: number; y: number; }} delta
 * @param {boolean} flying;
 */
function moveWithCollision(entity, delta, flying = false) {
  const rect = new PIXI.Rectangle(entity.pos.x - entity.collideBox.width / 2, entity.pos.y - entity.collideBox.height / 2, entity.collideBox.width, entity.collideBox.height);

  if (delta.x > 0) {
    for (let tileX = Math.floor(rect.right); tileX <= Math.ceil(rect.right + delta.x); tileX++) {
      for (let tileY = Math.floor(rect.top); tileY <= Math.floor(rect.bottom); tileY++) {
        const tileCollision = getTileCollisionAt(tileX, tileY);
        if (tileCollision == null || (flying && tileCollision.ignoreFlyers)) continue;

        if (rect.bottom >= tileCollision.rect.top && rect.top <= tileCollision.rect.bottom) {
          if (rect.left < tileCollision.rect.left) {
            const diff = tileCollision.rect.left - (rect.right + delta.x);
            if (diff < 0) delta.x += diff - 1 / tileSize;
          }
        }
      }
    }
  } else if (delta.x < 0) {
    for (let tileX = Math.floor(rect.left); tileX >= Math.floor(rect.left + delta.x); tileX--) {
      for (let tileY = Math.floor(rect.top); tileY <= Math.floor(rect.bottom); tileY++) {
        const tileCollision = getTileCollisionAt(tileX, tileY);
        if (tileCollision == null || (flying && tileCollision.ignoreFlyers)) continue;

        if (rect.bottom >= tileCollision.rect.top && rect.top <= tileCollision.rect.bottom) {
          if (rect.right > tileCollision.rect.right) {
            const diff = tileCollision.rect.right - (rect.left + delta.x);
            if (diff > 0) delta.x += diff + 1 / tileSize;
          }
        }
      }
    }
  }

  rect.x += delta.x;

  if (delta.y > 0) {
    for (let tileY = Math.floor(rect.bottom); tileY <= Math.ceil(rect.bottom + delta.y); tileY++) {
      for (let tileX = Math.floor(rect.left); tileX <= Math.floor(rect.right); tileX++) {
        const tileCollision = getTileCollisionAt(tileX, tileY);
        if (tileCollision == null || (flying && tileCollision.ignoreFlyers)) continue;

        if (rect.right >= tileCollision.rect.left && rect.left <= tileCollision.rect.right) {
          if (rect.top < tileCollision.rect.top) {
            const diff = tileCollision.rect.top - (rect.bottom + delta.y);
            if (diff < 0) delta.y += diff - 1 / tileSize;
          }
        }
      }
    }
  } else if (delta.y < 0) {
    for (let tileY = Math.floor(rect.top); tileY >= Math.floor(rect.top + delta.y); tileY--) {
      for (let tileX = Math.floor(rect.left); tileX <= Math.floor(rect.right); tileX++) {
        const tileCollision = getTileCollisionAt(tileX, tileY);
        if (tileCollision == null || (flying && tileCollision.ignoreFlyers)) continue;

        if (rect.right >= tileCollision.rect.left && rect.left <= tileCollision.rect.right) {
          if (rect.bottom > tileCollision.rect.bottom) {
            const diff = tileCollision.rect.bottom - (rect.top + delta.y);
            if (diff > 0) delta.y += diff + 1 / tileSize;
          }
        }
      }
    }
  }

  entity.pos.x += delta.x;
  entity.pos.y += delta.y;
}

/**
 * @param {EntityLike} a
 * @param {EntityLike} b
 * @param {number} dt
 * @param {boolean} moveB
 */
function pushaway(a, b, dt, moveB = true) {
  const diff = { x: a.pos.x - b.pos.x, y: a.pos.y - b.pos.y };
  const overlap = lengthxy(diff) - (a.collideBox.width + b.collideBox.width);
  const angle = Math.atan2(diff.y, diff.x);

  if (overlap < 0) {
    const amountA = Math.min(-overlap, a.speed * dt);
    moveWithCollision(a, { x: Math.cos(angle) * amountA, y: Math.sin(angle) * amountA });

    if (moveB) {
      const amountB = Math.min(-overlap, b.speed * dt);
      moveWithCollision(b, { x: -Math.cos(angle) * amountB, y: -Math.sin(angle) * amountB });
    }
  }
}
